package store

import (
	"encoding/json"
	"fmt"
	log "github.com/sirupsen/logrus"
	"os"
)

// User represents metadata of a user.
type User struct {
	Secret     string `json:"secret"`
	Privileged bool   `json:"privileged"`
	Snowflake  string `json:"snowflake"`
	Username   string `json:"username"`
}

// user parses user metadata file.
func (s *Store) user(path string) User {
	if payload, err := os.ReadFile(path); err != nil {
		s.fatalClose(fmt.Sprintf("Error reading user metadata %s, %s", path, err))
	} else {
		var info User
		if err = json.Unmarshal(payload, &info); err != nil {
			s.fatalClose(fmt.Sprintf("Error parsing user metadata %s, %s", path, err))
		} else {
			return info
		}
	}
	return User{}
}

// User returns user information with specific snowflake.
func (s *Store) User(flake string) User {
	if !numerical(flake) || !s.file(s.UserPath(flake)) {
		return User{}
	}

	s.getLock(flake).RLock()
	defer s.getLock(flake).RUnlock()
	return s.user(s.UserMetadataPath(flake))
}

// Users returns a slice of user snowflakes.
func (s *Store) Users() []string {
	var users []string
	if entries, err := os.ReadDir(s.UsersDir()); err != nil {
		s.fatalClose(fmt.Sprintf("Error reading users, %s", err))
	} else {
		for _, entry := range entries {
			if entry.IsDir() {
				users = append(users, entry.Name())
			}
		}
	}
	return users
}

// UserMetadata sets user metadata.
func (s *Store) UserMetadata(info User) {
	if payload, err := json.Marshal(info); err != nil {
		s.fatalClose(fmt.Sprintf("Error updating user %s metadata, %s", info.Snowflake, err))
	} else {
		if err = os.WriteFile(s.UserMetadataPath(info.Snowflake), payload, s.PermissionFile); err != nil {
			s.fatalClose(fmt.Sprintf("Error writing user %s metadata, %s", info.Snowflake, err))
		}
	}
}

// UserAdd creates a user.
func (s *Store) UserAdd(username, password string, privileged bool) User {
	if len(username) > 64 || !nameRegex.MatchString(username) || s.file(s.UsernamePath(username)) {
		return User{}
	}

	info := User{
		Secret:     s.SecretNew(),
		Privileged: privileged,
		Snowflake:  userNode.Generate().String(),
		Username:   username,
	}
	// Create user directory and images
	if err := os.MkdirAll(s.UserImagesPath(info.Snowflake), s.PermissionDir); err != nil {
		s.fatalClose(fmt.Sprintf("Error creating user %s directory, %s", info.Snowflake, err))
	}
	s.getLock(info.Snowflake).Lock()
	defer s.getLock(info.Snowflake).Unlock()
	// Save user metadata
	s.UserMetadata(info)
	// Associate new user secret
	s.SecretAssociate(info.Secret, info.Snowflake)
	// Associate username
	s.userUsernameAssociate(info.Snowflake, info.Username)
	// Set password
	s.userPasswordUpdate(s.UserPath(info.Snowflake), password)

	log.Infof("User %s added with username %s privilege %v secret %s.", info.Snowflake, info.Username, info.Privileged, info.Secret)
	return info
}

// UserPrivileged sets privileged status of user with specific snowflake.
func (s *Store) UserPrivileged(flake string, privileged bool) {
	if !numerical(flake) {
		return
	}

	s.getLock(flake).Lock()
	defer s.getLock(flake).Unlock()

	info := s.user(s.UserMetadataPath(flake))
	if info.Snowflake != flake {
		return
	}
	info.Privileged = privileged
	s.UserMetadata(info)
	log.Infof("User %s privileged %v", flake, privileged)
}

// UserUsernameUpdate updates username of user with specific snowflake.
func (s *Store) UserUsernameUpdate(flake, username string) bool {
	if !numerical(flake) || !nameRegex.MatchString(username) || s.file(s.UsernamePath(username)) {
		return false
	}

	s.getLock(flake).Lock()
	defer s.getLock(flake).Unlock()

	info := s.user(s.UserMetadataPath(flake))
	if info.Snowflake != flake {
		return false
	}
	// Lock old username
	s.getLock(info.Username).Lock()
	defer s.getLock(info.Username).Unlock()
	// Disassociate old username and associate new
	if info.Username != "" {
		s.userUsernameDisassociate(info.Username)
	}
	s.userUsernameAssociate(flake, username)
	// Set username in metadata
	info.Username = username
	s.UserMetadata(info)

	log.Infof("User %s username updated to %s.", flake, username)
	return true
}

// UserSecretRegen regenerates secret of user with specific snowflake.
func (s *Store) UserSecretRegen(flake string) string {
	if !numerical(flake) {
		return ""
	}

	s.getLock(flake).Lock()
	defer s.getLock(flake).Unlock()

	info := s.user(s.UserMetadataPath(flake))
	// Check if obtained info is valid
	if info.Snowflake != flake {
		return ""
	}
	// Disassociate old user
	s.SecretDisassociate(info.Secret)
	// Generate new secret
	info.Secret = s.SecretNew()
	// Write metadata
	s.UserMetadata(info)
	// Associate new secret
	s.SecretAssociate(info.Secret, info.Snowflake)
	// Log the event
	log.Infof("User %s secret reset to %s.", flake, info.Secret)
	return info.Secret
}

// UserUsername returns user via username.
func (s *Store) UserUsername(username string) User {
	if !nameRegex.MatchString(username) || !s.file(s.UsernamePath(username)) {
		return User{}
	}
	s.getLock(username).RLock()
	user := s.user(s.readlink(s.UsernamePath(username)) + "/" + infoJson)
	s.getLock(username).RUnlock()
	return user
}

// userUsernameAssociate associates user snowflake with specific username.
func (s *Store) userUsernameAssociate(flake, username string) {
	s.link("../users/"+flake, s.UsernamePath(username))
}

// userUsernameDisassociate disassociates specific username.
func (s *Store) userUsernameDisassociate(username string) {
	if err := os.Remove(s.UsernamePath(username)); err != nil {
		s.fatalClose(fmt.Sprintf("Error disassociating username %s, %s", username, err))
	}
}

// userPassword returns password of user from path to user directory.
func (s *Store) userPassword(path string) string {
	if payload, err := os.ReadFile(path + "/passwd"); err != nil {
		if os.IsNotExist(err) {
			return ""
		}
		s.fatalClose(fmt.Sprintf("Error reading password from user directory %s, %s", path, err))
	} else {
		return string(payload)
	}
	return ""
}

// userPasswordUpdate updates user password of user from path to user directory.
func (s *Store) userPasswordUpdate(path, password string) bool {
	if len(password) > 1024 {
		return false
	}
	if err := os.WriteFile(path+"/passwd", []byte(password), s.PermissionFile); err != nil {
		s.fatalClose(fmt.Sprintf("Error setting password for user directory %s, %s", path, err))
	} else {
		return true
	}
	return false
}

// UserPasswordValidate validates password of specified user.
func (s *Store) UserPasswordValidate(flake, password string) bool {
	if !numerical(flake) || !s.file(s.UserPath(flake)) {
		return false
	}

	s.getLock(flake).RLock()
	defer s.getLock(flake).RUnlock()

	// Check password is not empty and matches
	if password != "" && password == s.userPassword(s.UserPath(flake)) {
		return true
	}

	return false
}

// UserUsernamePasswordValidate validates password of specified user from username.
func (s *Store) UserUsernamePasswordValidate(username, password string) bool {
	if !nameRegex.MatchString(username) || !s.file(s.UsernamePath(username)) {
		return false
	}

	s.getLock(username).RLock()
	defer s.getLock(username).RUnlock()

	if password != "" && password == s.userPassword(s.readlink(s.UsernamePath(username))) {
		return true
	}

	return false
}

// UserPasswordUpdate updates password of specified user.
func (s *Store) UserPasswordUpdate(flake, password string) bool {
	if !numerical(flake) {
		return false
	}

	s.getLock(flake).Lock()
	defer s.getLock(flake).Unlock()

	return s.userPasswordUpdate(s.UserPath(flake), password)
}

// UserDestroy destroys a user with specific snowflake.
func (s *Store) UserDestroy(flake string) {
	if !numerical(flake) {
		return
	}
	if !s.dir(s.UserPath(flake)) {
		return
	}

	s.getLock(flake).Lock()
	defer s.getLock(flake).Unlock()

	info := s.User(flake)
	// Check if info correct
	if info.Snowflake != flake {
		return
	}
	// Disassociate secret
	s.SecretDisassociate(info.Snowflake)
	// Remove user data directory
	if err := os.RemoveAll(s.UserPath(flake)); err != nil {
		s.fatalClose(fmt.Sprintf("Error destroying user %s data directory, %s", flake, err))
	}
	log.Infof("User %s username %s destroyed.", info.Snowflake, info.Username)
}

// UserImages returns slice of a user's images.
func (s *Store) UserImages(flake string) []string {
	if !numerical(flake) {
		return nil
	}
	if !s.dir(s.UserImagesPath(flake)) {
		return nil
	}

	var images []string
	if entries, err := os.ReadDir(s.UserImagesPath(flake)); err != nil {
		s.fatalClose(fmt.Sprintf("Error reading user %s images, %s", flake, err))
	} else {
		for _, entry := range entries {
			images = append(images, entry.Name())
		}
	}
	return images
}
